from __future__ import annotations
from typing import Optional, Dict, Any, List
from deltacat.compute.converter.constants import (
    DEFAULT_CONVERTER_TASK_MAX_PARALLELISM,
)
from deltacat.constants import DEFAULT_NAMESPACE
from fsspec import AbstractFileSystem
from pyiceberg.catalog import Catalog


class ConverterSessionParams(dict):
    """
    This class represents the parameters passed to convert_ (deltacat/compute/compactor/compaction_session.py)
    """

    @staticmethod
    def of(params: Optional[Dict[str, Any]]) -> ConverterSessionParams:
        params = {} if params is None else params
        assert params.get("catalog") is not None, "catalog is a required arg"
        assert (
            params.get("iceberg_table_name") is not None
        ), "iceberg_table_name is a required arg"
        assert (
            params.get("iceberg_warehouse_bucket_name") is not None
        ), "iceberg_warehouse_bucket_name is a required arg"
        result = ConverterSessionParams(params)

        result.iceberg_namespace = params.get("iceberg_namespace", DEFAULT_NAMESPACE)
        result.enforce_primary_key_uniqueness = params.get(
            "enforce_primary_key_uniqueness", False
        )
        result.compact_previous_position_delete_files = params.get(
            "compact_previous_position_delete_files", False
        )

        # For Iceberg v3 spec, option to produce delete vector that can establish 1:1 mapping with data files.
        result.position_delete_for_multiple_data_files = params.get(
            "position_delete_for_multiple_data_files", True
        )
        result.task_max_parallelism = params.get(
            "task_max_parallelism", DEFAULT_CONVERTER_TASK_MAX_PARALLELISM
        )
        result.merge_keys = params.get("merge_keys", None)
        result.s3_client_kwargs = params.get("s3_client_kwargs", {})
        result.filesystem = params.get("filesystem", None)
        result.s3_prefix_override = params.get("s3_prefix_override", None)

        return result

    @property
    def catalog(self) -> Catalog:
        return self["catalog"]

    @property
    def iceberg_table_name(self) -> str:
        return self["iceberg_table_name"]

    @property
    def iceberg_warehouse_bucket_name(self) -> str:
        return self["iceberg_warehouse_bucket_name"]

    @property
    def iceberg_namespace(self) -> str:
        return self["iceberg_namespace"]

    @iceberg_namespace.setter
    def iceberg_namespace(self, iceberg_namespace: str) -> None:
        self["iceberg_namespace"] = iceberg_namespace

    @property
    def enforce_primary_key_uniqueness(self) -> bool:
        return self["enforce_primary_key_uniqueness"]

    @enforce_primary_key_uniqueness.setter
    def enforce_primary_key_uniqueness(
        self, enforce_primary_key_uniqueness: bool
    ) -> None:
        self["enforce_primary_key_uniqueness"] = enforce_primary_key_uniqueness

    @property
    def compact_previous_position_delete_files(self) -> bool:
        return self["compact_previous_position_delete_files"]

    @compact_previous_position_delete_files.setter
    def compact_previous_position_delete_files(
        self, compact_previous_position_delete_files: bool
    ) -> None:
        self[
            "compact_previous_position_delete_files"
        ] = compact_previous_position_delete_files

    @property
    def position_delete_for_multiple_data_files(self) -> bool:
        return self["position_delete_for_multiple_data_files"]

    @position_delete_for_multiple_data_files.setter
    def position_delete_for_multiple_data_files(
        self, position_delete_for_multiple_data_files: bool
    ) -> None:
        self[
            "position_delete_for_multiple_data_files"
        ] = position_delete_for_multiple_data_files

    @property
    def task_max_parallelism(self) -> int:
        return self["task_max_parallelism"]

    @task_max_parallelism.setter
    def task_max_parallelism(self, task_max_parallelism: int) -> None:
        self["task_max_parallelism"] = task_max_parallelism

    @property
    def merge_keys(self) -> Optional[List[str]]:
        return self["merge_keys"]

    @merge_keys.setter
    def merge_keys(self, merge_keys: Optional[List[str]]) -> None:
        self["merge_keys"] = merge_keys

    @property
    def s3_client_kwargs(self) -> Dict[str, Any]:
        return self["s3_client_kwargs"]

    @s3_client_kwargs.setter
    def s3_client_kwargs(self, s3_client_kwargs: Dict[str, Any]) -> None:
        self["s3_client_kwargs"] = s3_client_kwargs

    @property
    def filesystem(self) -> Optional[AbstractFileSystem]:
        return self["filesystem"]

    @filesystem.setter
    def filesystem(self, filesystem: Optional[AbstractFileSystem]) -> None:
        self["filesystem"] = filesystem

    @property
    def location_provider_prefix_override(self) -> Optional[str]:
        return self["location_provider_prefix_override"]

    @location_provider_prefix_override.setter
    def location_provider_prefix_override(
        self, location_provider_prefix_override: Optional[str]
    ) -> None:
        self["location_provider_prefix_override"] = location_provider_prefix_override



from deltacat.compute.jobs.client import (
    DeltaCatJobClient,
    job_client,
    local_job_client,
)

from ray.job_submission import JobStatus

__all__ = [
    "job_client",
    "local_job_client",
    "DeltaCatJobClient",
    "JobStatus",
]



from __future__ import annotations

from typing import Optional

import pyarrow.ipc
from pyarrow import RecordBatch, RecordBatchFileReader

from deltacat.experimental.storage.rivulet.fs.file_provider import FileProvider
from deltacat.experimental.storage.rivulet.metastore.sst import SSTableRow
from deltacat.experimental.storage.rivulet.reader.data_reader import (
    RowAndKey,
    FileReader,
    FILE_FORMAT,
)
from deltacat.experimental.storage.rivulet.reader.pyarrow_data_reader import (
    RecordBatchRowIndex,
)
from deltacat.experimental.storage.rivulet.schema.schema import Schema


class FeatherFileReader(FileReader[RecordBatchRowIndex]):
    """
    Feather file reader. This class is not thread safe

    This is mostly a copy-pasta from ParquetFileReader
    TODO can consider abstracting code between this and ParquetFileReader
    """

    def __init__(
        self,
        sst_row: SSTableRow,
        file_provider: FileProvider,
        primary_key: str,
        schema: Schema,
    ):
        self.sst_row = sst_row
        self.input = file_provider.provide_input_file(sst_row.uri)

        self.key = primary_key
        self.feather_file = sst_row.uri

        self.schema = schema

        # Iterator from pyarrow iter_batches API call. Pyarrow manages state of traversal within parquet row groups

        """
        These variables keep state about where the iterator is current at. They are initialized in __enter__()
        """
        self._curr_batch: RecordBatch | None = None
        self._feather_reader: RecordBatchFileReader | None = None
        # Arrow only lets you read feather files chunk by chunk using
        # RecordBatchFileReader.get_batch(index)
        self._curr_batch_index = 0
        self._curr_row_offset = 0
        self._pk_col = None

    def peek(self) -> Optional[RowAndKey[FILE_FORMAT]]:
        """
        Peek next record

        Note that there is an edge case where peek() is called on the bounary between record batches
        This only happens curr_row_offset == curr_batch.num_rows, meaning next() or peek() would need to advance
        to the next record batch. When this happens, peek() increments _curr_batch and sets _curr_row_offset to 0

        :return: Optional of RowAndPrimaryKey
        """
        if not self.__is_initialized():
            raise RuntimeError(
                "ParquetFileReader must be initialized with __enter__ before reading"
            )

        if self.__need_to_advance_record_batch():
            try:
                self.__advance_record_batch()
            except StopIteration:
                return None

        pk = self._pk_col[self._curr_row_offset].as_py()
        return RowAndKey(
            RecordBatchRowIndex(self._curr_batch, self._curr_row_offset), pk
        )

    def __next__(self) -> RowAndKey[FILE_FORMAT]:
        if not self.__is_initialized():
            raise RuntimeError(
                "ParquetFileReader must be initialized with __enter__ before reading"
            )

        if self.__need_to_advance_record_batch():
            self.__advance_record_batch()
            pk = self._pk_col[0].as_py()
            return RowAndKey(RecordBatchRowIndex(self._curr_batch, 0), pk)
        else:
            pk = self._pk_col[self._curr_row_offset].as_py()
            offset = self._curr_row_offset
            self._curr_row_offset += 1
            return RowAndKey(RecordBatchRowIndex(self._curr_batch, offset), pk)

    def __enter__(self):
        with self.input.open() as f:
            self._feather_reader = pyarrow.ipc.RecordBatchFileReader(f)
            self.__advance_record_batch()

    def __exit__(self, __exc_type, __exc_value, __traceback):
        self.close()
        # return False to propagate up error messages
        return False

    def close(self):
        # no op
        return

    def __is_initialized(self):
        return self._curr_batch and self._pk_col

    def __need_to_advance_record_batch(self):
        return not self._curr_row_offset < self._curr_batch.num_rows

    def __advance_record_batch(self):
        """
        Advance to next record batch
        :raise StopIteration: If there are no more record batches
        """
        try:
            self._curr_batch = self._feather_reader.get_batch(self._curr_batch_index)
            self._curr_batch_index += 1
            self._curr_row_offset = 0
            self._pk_col = self._curr_batch[self.key]
            # Filter the batch to only include fields in the schema
            # Pyarrow select will throw a ValueError if the field is not in the schema
            fields = [
                field
                for field in self.schema.keys()
                if field in self._curr_batch.schema.names
            ]
            self._curr_batch = self._curr_batch.select(fields)
        except ValueError:
            raise StopIteration(f"Ended iteration at batch {self._curr_batch_index}")



from deltacat.constants import DEFAULT_NAMESPACE
from deltacat.utils.ray_utils.concurrency import (
    invoke_parallel,
    task_resource_options_provider,
)
import ray
import functools
from deltacat.compute.converter.utils.convert_task_options import (
    convert_resource_options_provider,
)
import logging
from deltacat import logs
from deltacat.compute.converter.model.converter_session_params import (
    ConverterSessionParams,
)
from typing import Dict, List, Any, Callable
from deltacat.compute.converter.constants import DEFAULT_MAX_PARALLEL_DATA_FILE_DOWNLOAD
from deltacat.compute.converter.steps.convert import convert
from deltacat.compute.converter.model.convert_input import ConvertInput
from deltacat.compute.converter.pyiceberg.overrides import (
    fetch_all_bucket_files,
)
from deltacat.compute.converter.utils.converter_session_utils import (
    construct_iceberg_table_prefix,
)
from deltacat.compute.converter.pyiceberg.update_snapshot_overrides import (
    commit_replace_snapshot,
    commit_append_snapshot,
)
from deltacat.compute.converter.pyiceberg.catalog import load_table
from deltacat.compute.converter.utils.converter_session_utils import (
    group_all_files_to_each_bucket,
)
from deltacat.compute.converter.model.convert_result import ConvertResult
from deltacat.compute.converter.utils.converter_session_utils import (
    _get_snapshot_action_description,
    _determine_snapshot_type,
    SnapshotType,
)

from pyiceberg.manifest import DataFile
from pyiceberg.table.metadata import TableMetadata

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def converter_session(params: ConverterSessionParams, **kwargs: Any) -> TableMetadata:
    """
    Convert equality deletes to position deletes with option to enforce primary key uniqueness.

    This function processes Iceberg table files to convert equality delete files to position delete files.
    It can optionally enforce primary key uniqueness by keeping only the latest version of each
    primary key across all data files.

    **Memory Requirements:**
    - Minimum 512MB of free memory is required to run the converter

    **Process Overview:**
    1. Fetches all bucket files (data files, equality deletes, position deletes)
    2. Groups files by bucket for parallel processing
    3. Converts equality deletes to position deletes using Ray parallel tasks
    4. Enforces primary key uniqueness if enabled
    5. Commits appropriate snapshot (append, replace, or delete) to the Iceberg table


    Args:
        params: ConverterSessionParams containing all configuration parameters
            - catalog: Iceberg catalog instance
            - iceberg_table_name: Name of the target Iceberg table
            - enforce_primary_key_uniqueness: Whether to enforce PK uniqueness
            - iceberg_warehouse_bucket_name: S3 bucket for Iceberg warehouse
            - iceberg_namespace: Iceberg namespace
            - merge_keys: Optional list of merge key fields (uses table identifier fields if not provided)
            - compact_previous_position_delete_files: Whether to compact existing position delete files
            - task_max_parallelism: Maximum number of parallel Ray tasks
            - s3_client_kwargs: Additional S3 client configuration
            - s3_file_system: S3 file system instance
            - location_provider_prefix_override: Optional prefix override for file locations
            - position_delete_for_multiple_data_files: Whether to generate position deletes for multiple data files
        **kwargs: Additional keyword arguments (currently unused)

    Raises:
        Exception: If snapshot commitment fails or other critical errors occur

    """

    catalog = params.catalog
    table_name = params.iceberg_table_name
    if "." not in table_name:
        iceberg_namespace = params.iceberg_namespace or DEFAULT_NAMESPACE
        table_name = params.iceberg_table_name
        table_identifier = f"{iceberg_namespace}.{table_name}"
    else:
        table_identifier = table_name
        identifier_parts = table_identifier.split(".")
        iceberg_namespace = identifier_parts[0]
        table_name = identifier_parts[1]
    iceberg_table = load_table(catalog, table_identifier)
    enforce_primary_key_uniqueness = params.enforce_primary_key_uniqueness
    iceberg_warehouse_bucket_name = params.iceberg_warehouse_bucket_name
    merge_keys = params.merge_keys
    compact_previous_position_delete_files = (
        params.compact_previous_position_delete_files
    )
    task_max_parallelism = params.task_max_parallelism
    s3_client_kwargs = params.s3_client_kwargs
    s3_file_system = params.filesystem
    location_provider_prefix_override = params.location_provider_prefix_override
    position_delete_for_multiple_data_files = (
        params.position_delete_for_multiple_data_files
    )

    data_file_dict, equality_delete_dict, pos_delete_dict = fetch_all_bucket_files(
        iceberg_table
    )

    convert_input_files_for_all_buckets = group_all_files_to_each_bucket(
        data_file_dict=data_file_dict,
        equality_delete_dict=equality_delete_dict,
        pos_delete_dict=pos_delete_dict,
    )

    if not location_provider_prefix_override:
        iceberg_table_warehouse_prefix = construct_iceberg_table_prefix(
            iceberg_warehouse_bucket_name=iceberg_warehouse_bucket_name,
            table_name=table_name,
            iceberg_namespace=iceberg_namespace,
        )
    else:
        iceberg_table_warehouse_prefix = location_provider_prefix_override

    # Using table identifier fields as merge keys if merge keys not provided
    if not merge_keys:
        identifier_fields_set = iceberg_table.schema().identifier_field_names()
        identifier_fields = list(identifier_fields_set)
    else:
        identifier_fields = merge_keys

    convert_options_provider: Callable = functools.partial(
        task_resource_options_provider,
        resource_amount_provider=convert_resource_options_provider,
    )

    # TODO (zyiqin): max_parallel_data_file_download should be determined by memory requirement for each bucket.
    #  Specifically, for case when files for one bucket memory requirement exceed one worker node's memory limit, WITHOUT rebasing with larger hash bucket count,
    #  1. We can control parallel files to download by adjusting max_parallel_data_file_download.
    #  2. Implement two-layer converter tasks, with convert tasks to spin up child convert tasks.
    #  Note that approach 2 will ideally require shared object store to avoid download equality delete files * number of child tasks times.
    max_parallel_data_file_download = DEFAULT_MAX_PARALLEL_DATA_FILE_DOWNLOAD

    def convert_input_provider(index: int, item: Any) -> Dict[str, ConvertInput]:
        task_opts = convert_options_provider(index, item)
        return {
            "convert_input": ConvertInput.of(
                convert_input_files=item,
                convert_task_index=index,
                iceberg_table_warehouse_prefix=iceberg_table_warehouse_prefix,
                identifier_fields=identifier_fields,
                compact_previous_position_delete_files=compact_previous_position_delete_files,
                table_io=iceberg_table.io,
                table_metadata=iceberg_table.metadata,
                enforce_primary_key_uniqueness=enforce_primary_key_uniqueness,
                position_delete_for_multiple_data_files=position_delete_for_multiple_data_files,
                max_parallel_data_file_download=max_parallel_data_file_download,
                s3_client_kwargs=s3_client_kwargs,
                filesystem=s3_file_system,
                task_memory=task_opts["memory"],
            )
        }

    logger.info(f"Getting remote convert tasks...")
    # Ray remote task: convert
    # TODO: Add split mechanism to split large buckets
    convert_tasks_pending = invoke_parallel(
        items=convert_input_files_for_all_buckets,
        ray_task=convert,
        max_parallelism=task_max_parallelism,
        options_provider=convert_options_provider,
        kwargs_provider=convert_input_provider,
    )

    to_be_deleted_files_list: List[List[DataFile]] = []
    logger.info(f"Finished invoking {len(convert_tasks_pending)} convert tasks.")

    convert_results: List[ConvertResult] = ray.get(convert_tasks_pending)
    logger.info(f"Got {len(convert_tasks_pending)} convert tasks.")

    total_position_delete_record_count = sum(
        convert_result.position_delete_record_count
        for convert_result in convert_results
    )
    total_input_data_file_record_count = sum(
        convert_result.input_data_files_record_count
        for convert_result in convert_results
    )
    total_data_file_hash_columns_in_memory_sizes = sum(
        convert_result.input_data_files_hash_columns_in_memory_sizes
        for convert_result in convert_results
    )
    total_position_delete_file_in_memory_sizes = sum(
        convert_result.position_delete_in_memory_sizes
        for convert_result in convert_results
    )
    total_position_delete_on_disk_sizes = sum(
        convert_result.position_delete_on_disk_sizes
        for convert_result in convert_results
    )
    total_input_data_files_on_disk_size = sum(
        convert_result.input_data_files_on_disk_size
        for convert_result in convert_results
    )

    # Calculate memory usage statistics
    max_peak_memory_usage = max(
        convert_result.peak_memory_usage_bytes for convert_result in convert_results
    )
    avg_memory_usage_percentage = sum(
        convert_result.memory_usage_percentage for convert_result in convert_results
    ) / len(convert_results)
    max_memory_usage_percentage = max(
        convert_result.memory_usage_percentage for convert_result in convert_results
    )

    logger.info(
        f"Aggregated stats for {table_identifier}: "
        f"total position delete record count: {total_position_delete_record_count}, "
        f"total input data file record count: {total_input_data_file_record_count}, "
        f"total data file hash columns in memory sizes: {total_data_file_hash_columns_in_memory_sizes}, "
        f"total position delete file in memory sizes: {total_position_delete_file_in_memory_sizes}, "
        f"total position delete file on disk sizes: {total_position_delete_on_disk_sizes}, "
        f"total input data files on disk size: {total_input_data_files_on_disk_size}, "
        f"max peak memory usage: {max_peak_memory_usage} bytes, "
        f"average memory usage percentage: {avg_memory_usage_percentage:.2f}%, "
        f"max memory usage percentage: {max_memory_usage_percentage:.2f}%"
    )

    to_be_added_files_list: List[DataFile] = []
    for convert_result in convert_results:
        to_be_added_files = convert_result.to_be_added_files
        to_be_deleted_files = convert_result.to_be_deleted_files

        to_be_deleted_files_list.extend(to_be_deleted_files.values())
        to_be_added_files_list.extend(to_be_added_files)

    logger.info(f"To be deleted files list length: {len(to_be_deleted_files_list)}")
    logger.info(f"To be added files list length: {len(to_be_added_files_list)}")

    # Determine snapshot type and commit
    snapshot_type = _determine_snapshot_type(
        to_be_deleted_files_list, to_be_added_files_list
    )

    if snapshot_type == SnapshotType.NONE:
        logger.info(
            _get_snapshot_action_description(
                snapshot_type, to_be_deleted_files_list, to_be_added_files_list
            )
        )
        return

    logger.info(
        f"Snapshot action: {_get_snapshot_action_description(snapshot_type, to_be_deleted_files_list, to_be_added_files_list)}"
    )

    try:
        if snapshot_type == SnapshotType.APPEND:
            logger.info(f"Committing append snapshot for {table_identifier}.")
            updated_table_metadata = commit_append_snapshot(
                iceberg_table=iceberg_table,
                new_position_delete_files=to_be_added_files_list,
            )
        elif snapshot_type == SnapshotType.REPLACE:
            logger.info(f"Committing replace snapshot for {table_identifier}.")
            updated_table_metadata = commit_replace_snapshot(
                iceberg_table=iceberg_table,
                to_be_deleted_files=to_be_deleted_files_list,
                new_position_delete_files=to_be_added_files_list,
            )
        elif snapshot_type == SnapshotType.DELETE:
            logger.info(f"Committing delete snapshot for {table_identifier}.")
            updated_table_metadata = commit_replace_snapshot(
                iceberg_table=iceberg_table,
                to_be_deleted_files=to_be_deleted_files_list,
                new_position_delete_files=[],  # No new files to add
            )
        else:
            logger.warning(f"Unexpected snapshot type: {snapshot_type}")
            return

        logger.info(
            f"Committed new Iceberg snapshot for {table_identifier}: {updated_table_metadata.current_snapshot_id}"
        )

        # Return the updated table metadata with the new snapshot
        return updated_table_metadata
    except Exception as e:
        logger.error(f"Failed to commit snapshot for {table_identifier}: {str(e)}")
        raise



from collections import defaultdict
import logging
from deltacat import logs
import pyarrow
import pyarrow.parquet as pq
from pyiceberg.io.pyarrow import (
    parquet_path_to_id_mapping,
    StatisticsCollector,
    MetricModeTypes,
    DataFileStatistics,
    MetricsMode,
    StatsAggregator,
)
from typing import Dict, List, Set, Any, Tuple
from deltacat.compute.converter.utils.iceberg_columns import (
    ICEBERG_RESERVED_FIELD_ID_FOR_FILE_PATH_COLUMN,
    ICEBERG_RESERVED_FIELD_ID_FOR_POS_COLUMN,
)
from pyiceberg.io.pyarrow import (
    compute_statistics_plan,
)
from pyiceberg.manifest import (
    DataFile,
    DataFileContent,
    FileFormat,
)
from pyiceberg.table import _min_sequence_number, _open_manifest, Table
from pyiceberg.utils.concurrent import ExecutorFactory
from itertools import chain
from pyiceberg.typedef import (
    KeyDefaultDict,
)
from pyiceberg.schema import Schema
from pyiceberg.io import FileIO
from deltacat.compute.converter.model.convert_input_files import (
    DataFileList,
)


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def parquet_path_to_id_mapping_override(schema: Schema) -> Dict[str, int]:
    res = parquet_path_to_id_mapping(schema)
    # Override here to insert position delete reserved column field IDs
    res["file_path"] = ICEBERG_RESERVED_FIELD_ID_FOR_FILE_PATH_COLUMN
    res["pos"] = ICEBERG_RESERVED_FIELD_ID_FOR_POS_COLUMN
    return res


def data_file_statistics_from_parquet_metadata(
    parquet_metadata: pq.FileMetaData,
    stats_columns: Dict[int, StatisticsCollector],
    parquet_column_mapping: Dict[str, int],
) -> DataFileStatistics:
    """
    Overrides original Pyiceberg function: Compute and return DataFileStatistics that includes the following.

    - record_count
    - column_sizes
    - value_counts
    - null_value_counts
    - nan_value_counts
    - column_aggregates
    - split_offsets

    Args:
        parquet_metadata (pyarrow.parquet.FileMetaData): A pyarrow metadata object.
        stats_columns (Dict[int, StatisticsCollector]): The statistics gathering plan. It is required to
            set the mode for column metrics collection
        parquet_column_mapping (Dict[str, int]): The mapping of the parquet file name to the field ID
    """
    column_sizes: Dict[int, int] = {}
    value_counts: Dict[int, int] = {}
    split_offsets: List[int] = []

    null_value_counts: Dict[int, int] = {}
    nan_value_counts: Dict[int, int] = {}

    col_aggs = {}

    invalidate_col: Set[int] = set()
    for r in range(parquet_metadata.num_row_groups):
        # References:
        # https://github.com/apache/iceberg/blob/fc381a81a1fdb8f51a0637ca27cd30673bd7aad3/parquet/src/main/java/org/apache/iceberg/parquet/ParquetUtil.java#L232
        # https://github.com/apache/parquet-mr/blob/ac29db4611f86a07cc6877b416aa4b183e09b353/parquet-hadoop/src/main/java/org/apache/parquet/hadoop/metadata/ColumnChunkMetaData.java#L184

        row_group = parquet_metadata.row_group(r)

        data_offset = row_group.column(0).data_page_offset
        dictionary_offset = row_group.column(0).dictionary_page_offset

        if row_group.column(0).has_dictionary_page and dictionary_offset < data_offset:
            split_offsets.append(dictionary_offset)
        else:
            split_offsets.append(data_offset)

        for pos in range(parquet_metadata.num_columns):
            column = row_group.column(pos)
            field_id = parquet_column_mapping[column.path_in_schema]
            if field_id in stats_columns:
                stats_col = stats_columns[field_id]

                column_sizes.setdefault(field_id, 0)
                column_sizes[field_id] += column.total_compressed_size

                if stats_col.mode == MetricsMode(MetricModeTypes.NONE):
                    continue

                value_counts[field_id] = (
                    value_counts.get(field_id, 0) + column.num_values
                )

                if column.is_stats_set:
                    try:
                        statistics = column.statistics

                        if statistics.has_null_count:
                            null_value_counts[field_id] = (
                                null_value_counts.get(field_id, 0)
                                + statistics.null_count
                            )

                        if stats_col.mode == MetricsMode(MetricModeTypes.COUNTS):
                            continue

                        if field_id not in col_aggs:
                            col_aggs[field_id] = StatsAggregator(
                                stats_col.iceberg_type,
                                statistics.physical_type,
                                stats_col.mode.length,
                            )

                        col_aggs[field_id].update_min(statistics.min)
                        col_aggs[field_id].update_max(statistics.max)

                    except pyarrow.lib.ArrowNotImplementedError as e:
                        invalidate_col.add(field_id)
                        logger.warning(e)
            else:
                # Note: Removed original adding columns without stats to invalid column logic here
                logger.warning(
                    "PyArrow statistics missing for column %d when writing file", pos
                )

    split_offsets.sort()

    for field_id in invalidate_col:
        del col_aggs[field_id]
        del null_value_counts[field_id]

    return DataFileStatistics(
        record_count=parquet_metadata.num_rows,
        column_sizes=column_sizes,
        value_counts=value_counts,
        null_value_counts=null_value_counts,
        nan_value_counts=nan_value_counts,
        column_aggregates=col_aggs,
        split_offsets=split_offsets,
    )


def parquet_files_dict_to_iceberg_data_files(
    io: FileIO,
    table_metadata: Any,
    files_dict: Dict[Any, List[str]],
    file_content_type: DataFileContent,
) -> List[DataFile]:
    iceberg_files = []
    schema = table_metadata.schema()
    for partition_value, file_paths in files_dict.items():
        for file_path in file_paths:
            input_file = io.new_input(file_path)
            with input_file.open() as input_stream:
                parquet_metadata = pq.read_metadata(input_stream)

            # Removed _check_pyarrow_schema_compatible() here since reserved columns does not comply to all rules.

            statistics = data_file_statistics_from_parquet_metadata(
                parquet_metadata=parquet_metadata,
                stats_columns=compute_statistics_plan(
                    schema, table_metadata.properties
                ),
                parquet_column_mapping=parquet_path_to_id_mapping_override(schema),
            )

            data_file = DataFile(
                content=file_content_type,
                file_path=file_path,
                file_format=FileFormat.PARQUET,
                partition=partition_value,
                file_size_in_bytes=len(input_file),
                sort_order_id=None,
                spec_id=table_metadata.default_spec_id,
                equality_ids=None,
                key_metadata=None,
                **statistics.to_serialized_dict(),
            )
            iceberg_files.append(data_file)
    return iceberg_files


def fetch_all_bucket_files(
    table: Table,
) -> Tuple[Dict[Any, DataFileList], Dict[Any, DataFileList], Dict[Any, DataFileList]]:
    # step 1: filter manifests using partition summaries
    # the filter depends on the partition spec used to write the manifest file, so create a cache of filters for each spec id
    data_scan = table.scan()
    snapshot = data_scan.snapshot()
    if not snapshot:
        return iter([])
    manifest_evaluators = KeyDefaultDict(data_scan._build_manifest_evaluator)

    manifests = [
        manifest_file
        for manifest_file in snapshot.manifests(data_scan.io)
        if manifest_evaluators[manifest_file.partition_spec_id](manifest_file)
    ]

    # step 2: filter the data files in each manifest
    # this filter depends on the partition spec used to write the manifest file
    partition_evaluators = KeyDefaultDict(data_scan._build_partition_evaluator)
    residual_evaluators = KeyDefaultDict(data_scan._build_residual_evaluator)
    min_sequence_number = _min_sequence_number(manifests)

    # {"bucket_index": List[DataFile]}
    data_entries = defaultdict(list)
    equality_data_entries = defaultdict(list)
    positional_delete_entries = defaultdict(list)

    executor = ExecutorFactory.get_or_create()
    for manifest_entry in chain(
        *executor.map(
            lambda args: _open_manifest(*args),
            [
                (
                    data_scan.io,
                    manifest,
                    partition_evaluators[manifest.partition_spec_id],
                    residual_evaluators[manifest.partition_spec_id],
                    data_scan._build_metrics_evaluator(),
                )
                for manifest in manifests
                if data_scan._check_sequence_number(min_sequence_number, manifest)
            ],
        )
    ):
        data_file = manifest_entry.data_file
        file_sequence_number = manifest_entry.sequence_number
        data_file_tuple = (file_sequence_number, data_file)
        partition_value = data_file.partition

        if data_file.content == DataFileContent.DATA:
            data_entries[partition_value].append(data_file_tuple)
        elif data_file.content == DataFileContent.POSITION_DELETES:
            positional_delete_entries[partition_value].append(data_file_tuple)
        elif data_file.content == DataFileContent.EQUALITY_DELETES:
            equality_data_entries[partition_value].append(data_file_tuple)
        else:
            logger.warning(
                f"Unknown DataFileContent ({data_file.content}): {manifest_entry}"
            )
    return data_entries, equality_data_entries, positional_delete_entries



from typing import Optional, Dict, List, Union, Tuple

import pyarrow as pa
from pyiceberg.catalog.rest import NAMESPACE_SEPARATOR

from pyiceberg.io import load_file_io
from pyiceberg.io.pyarrow import pyarrow_to_schema, schema_to_pyarrow
from pyiceberg.catalog import Catalog
from pyiceberg.partitioning import PartitionField, PartitionSpec
from pyiceberg.schema import (
    INITIAL_SCHEMA_ID,
    NestedField,
    Schema as IcebergSchema,
)
from pyiceberg.serializers import FromInputFile
from pyiceberg.table import (
    Table as IcebergTable,
    Namespace as IcebergNamespace,
    TableIdentifier,
)
from pyiceberg.table.metadata import TableMetadata
from pyiceberg.table.snapshots import MetadataLogEntry, Snapshot
from pyiceberg.table.sorting import (
    SortField,
    SortDirection,
    NullOrder as IcebergNullOrder,
    SortOrder as IcebergSortOrder,
)
from pyiceberg.transforms import (
    BucketTransform as IcebergBucketTransform,
    HourTransform as IcebergHourTransform,
    DayTransform as IcebergDayTransform,
    MonthTransform as IcebergMonthTransform,
    YearTransform as IcebergYearTransform,
    IdentityTransform as IcebergIdentityTransform,
    TruncateTransform as IcebergTruncateTransform,
    VoidTransform as IcebergIcebergVoidTransform,
    UnknownTransform as IcebergUnknownTransform,
    Transform as IcebergTransform,
)
from pyiceberg.typedef import Identifier, EMPTY_DICT

from deltacat.exceptions import (
    NamespaceNotFoundError,
    TableVersionNotFoundError,
    StreamNotFoundError,
    TableNotFoundError,
)
from deltacat.storage import (
    BucketingStrategy,
    BucketTransform,
    BucketTransformParameters,
    DayTransform,
    Field,
    HourTransform,
    IdentityTransform,
    MonthTransform,
    Namespace,
    NamespaceLocator,
    Schema,
    StreamLocator,
    Stream,
    Table,
    TableLocator,
    TableVersion,
    TableVersionLocator,
    Transform,
    TransformName,
    TruncateStrategy,
    TruncateTransform,
    TruncateTransformParameters,
    UnknownTransform,
    VoidTransform,
    YearTransform,
    SortOrder,
    NullOrder,
)
from deltacat.storage.model.interop import ModelMapper, OneWayModelMapper
from deltacat.storage.model.partition import PartitionKey, PartitionScheme
from deltacat.storage.model.sort_key import (
    SortKey,
    SortScheme,
)
from deltacat.storage.model.types import StreamFormat, CommitState


def _get_snapshot_for_meta(
    meta: TableMetadata,
    snapshot_id: int,
) -> Snapshot:
    try:
        return next(s for s in meta.snapshots if s.snapshot_id == snapshot_id)
    except StopIteration as e:
        err_msg = f"No table snapshot with ID: {snapshot_id}"
        raise ValueError(err_msg) from e


def _resolve_stream_snapshot(
    meta: TableMetadata,
    snapshot_id: Optional[int],
) -> Snapshot:
    sid = snapshot_id if snapshot_id else meta.current_snapshot_id
    try:
        return _get_snapshot_for_meta(meta, sid)
    except ValueError as e:
        err_msg = f"No snapshot with timestamp: {sid}.\nTable Metadata: {meta}"
        raise StreamNotFoundError(err_msg) from e


def _get_metadata_for_timestamp(
    timestamp: int,
    meta_log: List[MetadataLogEntry],
    catalog_properties: Dict[str, str] = EMPTY_DICT,
) -> TableMetadata:
    try:
        meta_log_entry = next(
            entry for entry in meta_log if entry.timestamp_ms == timestamp
        )
    except StopIteration as e:
        err_msg = f"No table metadata log with timestamp: {timestamp}"
        raise ValueError(err_msg) from e
    io = load_file_io(
        properties=catalog_properties,
        location=meta_log_entry.metadata_file,
    )
    file = io.new_input(meta_log_entry.metadata_file)
    return FromInputFile.table_metadata(file)


def _resolve_table_version_metadata(
    table: Optional[IcebergTable],
    timestamp: Optional[int] = None,
    catalog_properties: Dict[str, str] = EMPTY_DICT,
) -> TableMetadata:
    try:
        latest = table.metadata
        return (
            _get_metadata_for_timestamp(
                timestamp,
                table.metadata.metadata_log,
                catalog_properties,
            )
            if timestamp is not None and timestamp != latest.last_updated_ms
            else latest
        )
    except ValueError as e:
        raise TableVersionNotFoundError(
            f"Table version `{timestamp}` not found."
        ) from e


def _resolve_table_version(
    meta: TableMetadata,
    timestamp: Optional[int] = None,
) -> int:
    try:
        return (
            next(
                entry.timestamp_ms
                for entry in meta.metadata_log
                if entry.timestamp_ms == timestamp
            )
            if timestamp
            else meta.last_updated_ms
        )
    except StopIteration as e:
        err_msg = f"Table version `{timestamp}` not found."
        raise TableVersionNotFoundError(err_msg) from e


def _get_current_schema_for_meta(meta: TableMetadata) -> IcebergSchema:
    schema_id = meta.current_schema_id
    try:
        return next(schema for schema in meta.schemas if schema.schema_id == schema_id)
    except StopIteration as e:
        err_msg = f"No table schema with ID: {schema_id}"
        raise ValueError(err_msg) from e


def _get_current_spec_for_meta(meta: TableMetadata) -> PartitionSpec:
    spec_id = meta.default_spec_id
    try:
        return next(spec for spec in meta.partition_specs if spec.spec_id == spec_id)
    except StopIteration as e:
        err_msg = f"No table partition spec with ID: {spec_id}"
        raise ValueError(err_msg) from e


def _get_current_sort_order_for_meta(meta: TableMetadata) -> SortOrder:
    sort_order_id = meta.default_sort_order_id
    try:
        return next(
            sort_order
            for sort_order in meta.sort_orders
            if sort_order.order_id == sort_order_id
        )
    except StopIteration as e:
        err_msg = f"No table sort order with ID: {sort_order_id}"
        raise ValueError(err_msg) from e


class TransformMapper(ModelMapper[IcebergTransform, Transform]):
    @staticmethod
    def map(
        obj: Optional[IcebergTransform],
        **kwargs,
    ) -> Optional[Transform]:
        if obj is None:
            return None
        if isinstance(obj, IcebergIdentityTransform):
            return IdentityTransform.of()
        if isinstance(obj, IcebergHourTransform):
            return HourTransform.of()
        if isinstance(obj, IcebergDayTransform):
            return DayTransform.of()
        if isinstance(obj, IcebergMonthTransform):
            return MonthTransform.of()
        if isinstance(obj, IcebergYearTransform):
            return YearTransform.of()
        if isinstance(obj, IcebergIcebergVoidTransform):
            return VoidTransform.of()
        if isinstance(obj, IcebergBucketTransform):
            return BucketTransform.of(
                BucketTransformParameters.of(
                    num_buckets=obj.num_buckets,
                    bucketing_strategy=BucketingStrategy.ICEBERG,
                ),
            )
        if isinstance(obj, IcebergTruncateTransform):
            return TruncateTransform.of(
                TruncateTransformParameters.of(
                    width=obj.width,
                    truncate_strategy=TruncateStrategy.ICEBERG,
                ),
            )
        return UnknownTransform.of()

    @staticmethod
    def unmap(
        obj: Optional[Transform],
        **kwargs,
    ) -> Optional[IcebergTransform]:
        if obj is None:
            return None
        if obj.name == TransformName.IDENTITY:
            return IcebergIdentityTransform()
        if obj.name == TransformName.HOUR:
            return IcebergHourTransform()
        if obj.name == TransformName.DAY:
            return IcebergDayTransform()
        if obj.name == TransformName.MONTH:
            return IcebergMonthTransform()
        if obj.name == TransformName.YEAR:
            return IcebergYearTransform()
        if obj.name == TransformName.VOID:
            return IcebergIcebergVoidTransform()
        if obj.name == TransformName.BUCKET:
            parameters = BucketTransformParameters(obj.parameters)
            strategy = parameters.bucketing_strategy
            if strategy == BucketingStrategy.ICEBERG:
                return IcebergBucketTransform(parameters.num_buckets)
            else:
                err_msg = f"Unsupported Iceberg Bucketing Strategy: {strategy}."
                raise ValueError(err_msg)
        if obj.name == TransformName.TRUNCATE:
            parameters = TruncateTransformParameters(obj.parameters)
            return IcebergTruncateTransform(parameters.width)
        return IcebergUnknownTransform(obj.name)


class PartitionKeyMapper(ModelMapper[PartitionField, PartitionKey]):
    @staticmethod
    def map(
        obj: Optional[PartitionField],
        schema: IcebergSchema = IcebergSchema(),
        **kwargs,
    ) -> Optional[PartitionKey]:
        if obj is None:
            return None
        if not schema:
            err_msg = "Schema is required for Partition Field conversion."
            raise ValueError(err_msg)
        field = schema.find_field(name_or_id=obj.source_id)
        return PartitionKey.of(
            key=[field.name],
            name=obj.name,
            field_id=obj.field_id,
            transform=TransformMapper.map(obj.transform),
            native_object=obj,
        )

    @staticmethod
    def unmap(
        obj: Optional[PartitionKey],
        schema: IcebergSchema = IcebergSchema(),
        case_sensitive: bool = True,
    ) -> Optional[PartitionField]:
        if obj is None:
            return None
        if not schema:
            err_msg = "Schema is required for Partition Key conversion."
            raise ValueError(err_msg)
        if len(obj.key) > 1:
            err_msg = f"Iceberg only supports transforming 1 partition field."
            raise ValueError(err_msg)
        field = schema.find_field(
            name_or_id=obj.key[0],
            case_sensitive=case_sensitive,
        )
        return PartitionField(
            source_id=field.field_id,
            field_id=obj.id if obj.id else None,
            transform=TransformMapper.unmap(obj.transform),
            name=obj.name,
        )


class PartitionSchemeMapper(ModelMapper[PartitionSpec, PartitionScheme]):
    @staticmethod
    def map(
        obj: Optional[PartitionSpec],
        schema: IcebergSchema = IcebergSchema(),
        name: Optional[str] = None,
    ) -> Optional[PartitionScheme]:
        if obj is None:
            return None
        elif not schema:
            err_msg = "Schema is required for Partition Spec conversion."
            raise ValueError(err_msg)
        keys = [PartitionKeyMapper.map(field, schema) for field in obj.fields] or None
        return PartitionScheme.of(
            keys=keys,
            name=name,
            scheme_id=str(obj.spec_id),
            native_object=obj,
        )

    @staticmethod
    def unmap(
        obj: Optional[PartitionScheme],
        schema: IcebergSchema = IcebergSchema(),
        case_sensitive: bool = True,
    ) -> Optional[PartitionSpec]:
        if obj is None:
            return None
        if not schema:
            err_msg = "Schema is required for Partition Scheme conversion."
            raise ValueError(err_msg)
        fields = [
            PartitionKeyMapper.unmap(key, schema, case_sensitive) for key in obj.keys
        ]
        return PartitionSpec(
            fields=fields,
            spec_id=int(obj.id),
        )


class SortKeyMapper(ModelMapper[SortField, SortKey]):
    @staticmethod
    def unmap(
        obj: Optional[SortKey],
        schema: IcebergSchema = IcebergSchema(),
        case_sensitive: bool = True,
    ) -> Optional[SortField]:
        if obj is None:
            return None
        if not schema:
            err_msg = "Schema is required for Sort Key conversion."
            raise ValueError(err_msg)
        if len(obj.key) > 1:
            err_msg = f"Iceberg only supports transforming 1 sort field."
            raise ValueError(err_msg)
        field = schema.find_field(
            name_or_id=obj.key[0],
            case_sensitive=case_sensitive,
        )
        direction = (
            SortDirection.ASC
            if obj.sort_order is SortOrder.ASCENDING
            else SortDirection.DESC
            if obj.sort_order is SortOrder.DESCENDING
            else None
        )
        null_order = (
            IcebergNullOrder.NULLS_FIRST
            if obj.null_order is NullOrder.AT_START
            else IcebergNullOrder.NULLS_LAST
            if obj.null_order is NullOrder.AT_END
            else None
        )
        return SortField(
            source_id=field.field_id,
            transform=TransformMapper.unmap(obj.transform),
            direction=direction,
            null_order=null_order,
        )

    @staticmethod
    def map(
        obj: Optional[SortField],
        schema: IcebergSchema = IcebergSchema(),
        **kwargs,
    ) -> Optional[SortKey]:
        if obj is None:
            return None
        if not schema:
            err_msg = "Schema is required for Sort Field conversion."
            raise ValueError(err_msg)
        field = schema.find_field(name_or_id=obj.source_id)
        return SortKey.of(
            key=[field.name],
            sort_order=SortOrder(obj.direction.value or "ascending"),
            null_order=NullOrder(obj.null_order.value or "first"),
            transform=TransformMapper.map(obj.transform),
            native_object=obj,
        )


class SortSchemeMapper(ModelMapper[IcebergSortOrder, SortScheme]):
    @staticmethod
    def map(
        obj: Optional[IcebergSortOrder],
        schema: IcebergSchema = IcebergSchema(),
        name: Optional[str] = None,
        id: Optional[str] = None,
    ) -> Optional[SortScheme]:
        if obj is None:
            return None
        elif not schema:
            err_msg = "Schema is required for Sort Order conversion."
            raise ValueError(err_msg)
        keys = [SortKeyMapper.map(field, schema) for field in obj.fields] or None
        return SortScheme.of(
            keys=keys,
            name=name,
            scheme_id=id,
            native_object=obj,
        )

    @staticmethod
    def unmap(
        obj: Optional[SortScheme],
        schema: IcebergSchema = IcebergSchema(),
        case_sensitive: bool = True,
    ) -> Optional[IcebergSortOrder]:
        if obj is None:
            return None
        if not schema:
            err_msg = "Schema is required for Sort Scheme conversion."
            raise ValueError(err_msg)
        fields = [SortKeyMapper.unmap(key, schema, case_sensitive) for key in obj]
        return IcebergSortOrder(fields=fields)


class SchemaMapper(ModelMapper[IcebergSchema, Schema]):
    @staticmethod
    def map(
        obj: Optional[IcebergSchema],
        stream_locator: Optional[StreamLocator] = None,
        **kwargs,
    ) -> Optional[Schema]:
        if obj is None:
            return None
        schema: pa.Schema = schema_to_pyarrow(obj)
        # use DeltaCAT fields to extract field IDs from PyArrow schema metadata
        fields = [Field.of(field) for field in schema]
        final_fields = []
        for field in fields:
            iceberg_field = obj.find_field(field.id)
            final_field = Field.of(
                field=field.arrow,
                field_id=field.id,
                is_merge_key=field.id in obj.identifier_field_ids,
                doc=iceberg_field.doc,
                past_default=iceberg_field.initial_default,
                future_default=iceberg_field.write_default,
                native_object=iceberg_field,
            )
            final_fields.append(final_field)
        # TODO(pdames): Traverse DeltaCAT schemas to find one already related
        #  to this Iceberg schema.
        return Schema.of(
            schema=final_fields,
            native_object=obj,
        )

    @staticmethod
    def unmap(
        obj: Optional[Schema], stream_locator: Optional[StreamLocator] = None, **kwargs
    ) -> Optional[IcebergSchema]:
        if obj is None:
            return None
        if isinstance(obj.arrow, pa.Schema):
            schema = pyarrow_to_schema(obj.arrow)
            final_fields = []
            for field in obj.field_ids_to_fields.values():
                iceberg_field = schema.find_field(field.id)
                final_field = NestedField(
                    field_id=iceberg_field.field_id,
                    name=iceberg_field.name,
                    field_type=iceberg_field.field_type,
                    required=iceberg_field.required,
                    doc=field.doc,
                    initial_default=field.past_default,
                    write_default=field.future_default,
                )
                final_fields.append(final_field)
            # TODO (pmingshi): this code was changed as a hack to get schema conversion working
            # it still needs more testing
            iceberg_schema = IcebergSchema(
                fields=final_fields,
                schema_id=INITIAL_SCHEMA_ID,
                # identifier_field_ids=obj.merge_keys,
                identifier_field_ids=[],
            )
        else:
            err_msg = (
                f"unsupported schema type: `{type(obj.arrow)}`. "
                f"expected schema type: {pa.Schema}"
            )
            raise TypeError(err_msg)
        return iceberg_schema


class NamespaceLocatorMapper(
    ModelMapper[Union[Identifier, IcebergNamespace], NamespaceLocator]
):
    @staticmethod
    def map(
        obj: Optional[Union[Identifier, IcebergNamespace]], **kwargs
    ) -> Optional[NamespaceLocator]:
        namespace = None
        if obj is None:
            return None
        elif isinstance(obj, IcebergNamespace):
            namespace = NAMESPACE_SEPARATOR.join(obj.namespace.root[1:])
        elif isinstance(obj, Tuple):
            # In Iceberg, Tuple identifiers are of the form (namespace) or (namespace, table)
            # In this case, just take the first element of the tuple
            namespace = obj[0]
        if not namespace:
            err_msg = f"No namespace in identifier: {obj}"
            raise NamespaceNotFoundError(err_msg)
        return NamespaceLocator.of(namespace)

    @staticmethod
    def unmap(obj: Optional[NamespaceLocator], **kwargs) -> Optional[Identifier]:
        if obj is None:
            return None
        return tuple(obj.namespace.split("."))


class NamespaceMapper(ModelMapper[Union[Identifier, IcebergNamespace], Namespace]):
    @staticmethod
    def map(
        obj: Optional[Union[Identifier, IcebergNamespace]], **kwargs
    ) -> Optional[Namespace]:
        if obj is None:
            return None
        locator = NamespaceLocatorMapper.map(obj)
        return Namespace.of(locator=locator, properties=None)

    @staticmethod
    def unmap(
        obj: Optional[Namespace],
        **kwargs,
    ) -> Optional[Identifier]:
        if obj is None:
            return None
        return NamespaceLocatorMapper.unmap(obj.locator)


class TableLocatorMapper(ModelMapper[Union[Identifier, TableIdentifier], TableLocator]):
    @staticmethod
    def map(
        obj: Optional[Union[Identifier, TableIdentifier]], **kwargs
    ) -> Optional[TableLocator]:
        if obj is None:
            return None
        namespace_locator = NamespaceLocatorMapper.map(obj)
        table_name = (
            obj.name
            if isinstance(obj, TableIdentifier)
            else Catalog.table_name_from(obj)
        )
        if not table_name:
            raise TableNotFoundError(f"No table name in identifier: {obj}")
        return TableLocator.of(namespace_locator, table_name)

    @staticmethod
    def unmap(
        obj: Optional[TableLocator], catalog_name: Optional[str] = None, **kwargs
    ) -> Optional[Union[Identifier, TableIdentifier]]:
        if obj is None:
            return None
        identifier = tuple(obj.namespace.split(".")) + (obj.table_name,)
        return identifier


class TableMapper(OneWayModelMapper[IcebergTable, Table]):
    @staticmethod
    def map(
        obj: Optional[IcebergTable],
        **kwargs,
    ) -> Optional[Table]:
        if obj is None:
            return None
        locator = TableLocatorMapper.map(obj.name())
        return Table.of(
            locator=locator,
            description=None,
            properties=None,
            native_object=obj,
        )


class TableVersionLocatorMapper(OneWayModelMapper[IcebergTable, TableVersionLocator]):
    @staticmethod
    def map(
        obj: Optional[IcebergTable], timestamp: Optional[int] = None, **kwargs
    ) -> Optional[TableVersionLocator]:
        if obj is None:
            return None
        table_version = _resolve_table_version(obj.metadata, timestamp)
        return TableVersionLocator.of(
            table_locator=TableLocatorMapper.map(obj.name()),
            table_version=str(table_version),
        )


class TableVersionMapper(OneWayModelMapper[IcebergTable, TableVersion]):
    @staticmethod
    def map(
        obj: Optional[IcebergTable],
        timestamp: Optional[int] = None,
        catalog_properties: Dict[str, str] = EMPTY_DICT,
        **kwargs,
    ) -> Optional[TableVersion]:
        if obj is None:
            return None
        metadata = _resolve_table_version_metadata(obj, timestamp, catalog_properties)
        schema = _get_current_schema_for_meta(metadata)
        partition_spec = _get_current_spec_for_meta(metadata)
        sort_order = _get_current_sort_order_for_meta(metadata)
        return TableVersion.of(
            locator=TableVersionLocatorMapper.map(obj, timestamp),
            schema=SchemaMapper.map(schema),
            partition_scheme=PartitionSchemeMapper.map(partition_spec, schema),
            description=None,
            properties=obj.properties,
            content_types=None,
            sort_scheme=SortSchemeMapper.map(sort_order, schema),
            native_object=metadata,
        )


class StreamLocatorMapper(OneWayModelMapper[IcebergTable, StreamLocator]):
    @staticmethod
    def map(
        obj: Optional[IcebergTable],
        metadata_timestamp: Optional[int] = None,
        snapshot_id: Optional[int] = None,
        catalog_properties: Dict[str, str] = EMPTY_DICT,
        **kwargs,
    ) -> Optional[StreamLocator]:
        if obj is None:
            return None
        metadata = _resolve_table_version_metadata(
            obj, metadata_timestamp, catalog_properties
        )
        snapshot = _resolve_stream_snapshot(metadata, snapshot_id)
        return StreamLocator.of(
            table_version_locator=TableVersionLocatorMapper.map(
                obj, metadata_timestamp
            ),
            stream_id=str(snapshot.snapshot_id),
            stream_format=StreamFormat.ICEBERG.value,
        )


class StreamMapper(OneWayModelMapper[IcebergTable, Stream]):
    @staticmethod
    def map(
        obj: Optional[IcebergTable],
        # TODO (pdames): infer state from Iceberg metadata?
        state: Optional[CommitState] = CommitState.COMMITTED,
        metadata_timestamp: Optional[int] = None,
        snapshot_id: Optional[int] = None,
        catalog_properties: Dict[str, str] = EMPTY_DICT,
        **kwargs,
    ) -> Optional[Stream]:
        if obj is None:
            return None
        metadata = _resolve_table_version_metadata(
            obj, metadata_timestamp, catalog_properties
        )
        if not metadata.snapshots:
            return Stream.of(locator=None, partition_scheme=None)
        snapshot = _resolve_stream_snapshot(metadata, snapshot_id)
        schema = _get_current_schema_for_meta(metadata)
        partition_spec = _get_current_spec_for_meta(metadata)
        parent_snapshot_str = (
            str(snapshot.parent_snapshot_id) if snapshot.parent_snapshot_id else None
        )
        return Stream.of(
            locator=StreamLocatorMapper.map(
                obj, metadata_timestamp, snapshot_id, catalog_properties
            ),
            partition_scheme=PartitionSchemeMapper.map(partition_spec, schema),
            state=state,
            previous_stream_id=parent_snapshot_str,
            native_object=snapshot,
        )



import json

import pytest

from deltacat.storage.model.manifest import Manifest, ManifestEntry


@pytest.fixture
def manifest_a():
    return """
        {
          "entries":[
            {
              "uri":"s3://test_bucket/file1.tsv.gz",
              "mandatory":true,
              "meta":{
                "record_count":0,
                "content_length":123,
                "source_content_length":0,
                "content_type":"application/x-amzn-unescaped-tsv",
                "content_encoding":"gzip"
              }
            },
            {
              "uri":"s3://test_bucket/file2.tsv.gz",
              "mandatory":true,
              "meta":{
                "record_count":0,
                "content_length":456,
                "source_content_length":0,
                "content_type":"application/x-amzn-unescaped-tsv",
                "content_encoding":"gzip"
              }
            }
          ],
          "meta":{
            "record_count":0,
            "content_length":579,
            "source_content_length":0,
            "content_type":"application/x-amzn-unescaped-tsv",
            "content_encoding":"gzip"
          },
          "id":"052f62c0-5082-4935-9937-18a705156123",
          "author":{
            "name":"Dave",
            "version":"1.0"
          }
        }
           """


@pytest.fixture
def manifest_no_author():
    return """
        {
          "entries":[
            {
              "uri":"s3://test_bucket/file1.tsv.gz",
              "mandatory":true,
              "meta":{
                "record_count":0,
                "content_length":123,
                "source_content_length":0,
                "content_type":"application/x-amzn-unescaped-tsv",
                "content_encoding":"gzip"
              }
            },
            {
              "uri":"s3://test_bucket/file2.tsv.gz",
              "mandatory":true,
              "meta":{
                "record_count":0,
                "content_length":456,
                "source_content_length":0,
                "content_type":"application/x-amzn-unescaped-tsv",
                "content_encoding":"gzip"
              }
            }
          ],
          "meta":{
            "record_count":0,
            "content_length":579,
            "source_content_length":0,
            "content_type":"application/x-amzn-unescaped-tsv",
            "content_encoding":"gzip"
          },
          "id":"052f62c0-5082-4935-9937-18a705156123"
        }
           """


@pytest.fixture()
def manifest_entry_no_meta():
    return """
            {
              "uri":"s3://test_bucket/file1.tsv.gz",
              "mandatory":true
            }
           """


def test_manifest_from_json(manifest_a):
    manifest = Manifest.from_json(manifest_a)

    assert manifest.entries is not None
    assert len(manifest.entries) == 2
    assert manifest.entries[0].uri == "s3://test_bucket/file1.tsv.gz"
    assert manifest.entries[0].meta.record_count == 0
    assert manifest.meta.content_length == 579
    assert manifest.author.name == "Dave"


def test_manifest_from_json_no_author(manifest_no_author):
    manifest = Manifest.from_json(manifest_no_author)

    assert manifest.entries is not None
    assert len(manifest.entries) == 2
    assert manifest.entries[0].uri == "s3://test_bucket/file1.tsv.gz"
    assert manifest.entries[0].meta is not None
    assert manifest.author is None


def test_manifest_entry_from_dict_no_meta(manifest_entry_no_meta):
    entry = ManifestEntry.from_dict(json.loads(manifest_entry_no_meta))

    assert entry is not None
    assert entry.meta is None
    assert entry.uri == "s3://test_bucket/file1.tsv.gz"
    assert entry.mandatory is True



from typing import List, Optional

import pyarrow as pa
from abc import ABC, abstractmethod

from typing import Tuple
from deltacat.compute.compactor_v2.deletes.delete_file_envelope import (
    DeleteFileEnvelope,
)


class DeleteStrategy(ABC):
    """
    Encapsulates a strategy for applying row-level deletes on tables during compaction

    This abstract base class defines the interface for applying delete operations
    on intermediate in-memory pyarrow tables during compaction. Concrete subclasses must implement the `apply_deletes` and
    `apply_many_deletes` methods, as well as the `name` property.

    Example:
        >>> class MyDeleteStrategy(DeleteStrategy):
        ...     @property
        ...     def name(self) -> str:
        ...         return "MyDeleteStrategy"
        ...
        ...     def apply_deletes(self, table: Optional[pa.Table], delete_file_envelope: DeleteFileEnvelope) -> ReturnTuple[pa.Table, int]:
        ...         # Implement delete logic here
        ...         pass
        ...
        ...     def apply_many_deletes(self, table: Optional[pa.Table], delete_file_envelopes: List[DeleteFileEnvelope]) -> ReturnTuple[pa.Table, int]:
        ...         # Implement delete logic here
        ...         pass
    """

    @property
    def name(self) -> str:
        """
        The name of the delete strategy.
        """
        pass

    @abstractmethod
    def apply_deletes(
        self,
        table: Optional[pa.Table],
        delete_file_envelope: DeleteFileEnvelope,
        *args,
        **kwargs,
    ) -> Tuple[pa.Table, int]:
        """
        Apply delete operations on the given table using the provided delete file envelope.

        Args:
            table (Optional[pa.Table]): The pyarrow table to apply deletes on.
            delete_file_envelope (DeleteFileEnvelope): The delete file envelope containing delete parameters.

        Returns:
            Tuple[pa.Table, int]: A tuple containing the updated Arrow table after applying deletes,
                and the number of rows deleted.
        """
        pass

    @abstractmethod
    def apply_many_deletes(
        self,
        table: Optional[pa.Table],
        delete_file_envelopes: List[DeleteFileEnvelope],
        *args,
        **kwargs,
    ) -> Tuple[pa.Table, int]:
        """
        Apply delete operations on the given table using all provided delete file envelopes.

        Args:
            table (Optional[pa.Table]): The Arrow table to apply deletes on.
            delete_file_envelopes (List[DeleteFileEnvelope]): A list of delete file envelopes containing delete parameters.

        Returns:
            Tuple[pa.Table, int]: A tuple containing the updated Arrow table after applying all deletes,
                and the total number of rows deleted.
        """
        pass



import pytest
import shutil
import tempfile

from deltacat.experimental.storage.rivulet.fs.file_store import FileStore
from deltacat.experimental.storage.rivulet.writer.memtable_dataset_writer import (
    MemtableDatasetWriter,
)
from ..test_utils import (
    write_mvp_table,
    mvp_table_to_record_batches,
    validate_with_full_scan,
    create_dataset_for_method,
    assert_data_file_extension,
)

MemtableDatasetWriter.MAX_ROW_SIZE = 100


class TestWriter:
    temp_dir = None
    file_store: FileStore

    @classmethod
    def setup_class(cls):
        cls.temp_dir = tempfile.mkdtemp()
        path, filesystem = FileStore.filesystem(cls.temp_dir)
        cls.file_store = FileStore(path, filesystem)

    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.temp_dir)
        pass

    def test_write_unsupported_data_type(self, ds1_dataset, ds1_schema):
        dataset = create_dataset_for_method(self.temp_dir)
        dataset.add_schema(ds1_schema, "ds1_schema")
        with dataset.writer("ds1_schema") as writer:
            with pytest.raises(ValueError):
                writer.write("a string")

    def test_write_pydict(self, ds1_dataset, ds1_schema):
        dataset = create_dataset_for_method(self.temp_dir)
        dataset.add_schema(ds1_schema, "ds1_schema")
        with dataset.writer("ds1_schema") as writer:
            write_mvp_table(writer, ds1_dataset)

        validate_with_full_scan(dataset, ds1_dataset, ds1_schema)

    def test_write_record_batch(self, ds1_dataset, ds1_schema):
        dataset = create_dataset_for_method(self.temp_dir)
        dataset.add_schema(ds1_schema, "ds1_schema")
        with dataset.writer("ds1_schema") as writer:
            record_batch = mvp_table_to_record_batches(ds1_dataset, ds1_schema)
            writer.write(record_batch)

        validate_with_full_scan(dataset, ds1_dataset, ds1_schema)

    def test_write_list_of_record_batch(self, ds1_dataset, ds1_schema):
        dataset = create_dataset_for_method(self.temp_dir)
        dataset.add_schema(ds1_schema, "ds1_schema")
        with dataset.writer("ds1_schema", "feather") as writer:
            record_batch = mvp_table_to_record_batches(ds1_dataset, ds1_schema)
            writer.write([record_batch])

        validate_with_full_scan(dataset, ds1_dataset, ds1_schema)
        assert_data_file_extension(dataset, ".feather")

    def test_write_feather(self, dataset_images_with_label):
        dataset = create_dataset_for_method(self.temp_dir)

        table, schema = dataset_images_with_label
        dataset.add_schema(schema, "schema")
        with dataset.writer("schema", "feather") as writer:
            record_batch = mvp_table_to_record_batches(table, schema)
            writer.write([record_batch])

        validate_with_full_scan(dataset, table, schema)
        assert_data_file_extension(dataset, "feather")



# Similar to daft's datatype, this is a big ole enum of all possible types
#   In the long term, this will have to be interoperable with pandas/daft/spark/parquet/iceberg/etc type systems
#   Our Spec will need to publish data type mappings, such as Iceberg's data type mappings: https://iceberg.apache.org/spec/#file-system-operations
#   It also has the unique responsibility of representing multi-modal (e.g. image) types
from dataclasses import dataclass
from typing import Optional

import pyarrow as pa


# OPEN QUESTIONS:
# Do we want to support the notion of logical vs physical type like parquet?

# TODO turn into an interface or otherwise allow pluggable datatypes
@dataclass(frozen=True)
class Datatype:
    type_name: str

    @property
    def subtype(self) -> Optional[str]:
        """
        Higher level formats like binary or image will have "subtype", such as image(jpg) or binary(np_array)
        TODO - Note that we are replacing this schema system with DeltaCat schema model, which supports extended/decorated pyarrow types
        For now going to do a super minimal/hacky implementation of types like binary and image, where
        :return: Subtype if it exists, or None
        """
        if not self.type_name.endswith(")"):
            return None
        if self.type_name.startswith("binary(") or self.type_name.startswith("image("):
            return self.type_name[self.type_name.find("(") + 1 : -1]
        return None

    @classmethod
    def binary(cls, binary_format):
        """
        :param binary_format:
        :return:
        """
        return cls(type_name=f"binary({binary_format})")

    @classmethod
    def image(cls, image_format):
        return cls(type_name=f"image({image_format})")

    @classmethod
    def string(cls):
        return cls(type_name="string")

    @classmethod
    def float(cls):
        return cls(type_name="float")

    @classmethod
    def int16(cls):
        return cls(type_name="int16")

    @classmethod
    def int32(cls):
        return cls(type_name="int32")

    @classmethod
    def int64(cls):
        return cls(type_name="int64")

    @classmethod
    def bool(cls):
        return cls(type_name="bool")

    @classmethod
    def from_pyarrow(cls, pa_type: pa.DataType) -> "Datatype":
        """
        Convert a pa type to a Rivulet Datatype.

        Args:
            pa_type: pa DataType to convert

        Returns:
            Datatype: Corresponding Rivulet Datatype

        Raises:
            ValueError: If the pa type is not supported
        """
        if pa.types.is_string(pa_type):
            return cls.string()
        elif pa.types.is_float64(pa_type):
            return cls.float()
        elif pa.types.is_int16(pa_type):
            return cls.int16()
        elif pa.types.is_int32(pa_type):
            return cls.int32()
        elif pa.types.is_int64(pa_type):
            return cls.int64()
        elif pa.types.is_boolean(pa_type):
            return cls.bool()
        elif pa.types.is_binary(pa_type):
            # TODO: Use pyarrow metadata on schema field to map correctly into image and other binary types
            return cls.binary("binary")  # Default binary format
        else:
            raise ValueError(f"Unsupported pa type: {pa_type}")

    def to_pyarrow(self) -> pa.field:
        """
        In the future we want to be more thoughtful about how we do type conversions

        For now, just build a simple mapping of every time to pyarrow
        For what it's worth, Daft schema types have a giant if/else like this

        :return: pyarrow type
        """
        if self.type_name == "string":
            return pa.string()
        elif self.type_name == "float":
            return pa.float64()
        elif self.type_name == "int16":
            return pa.int16()
        elif self.type_name == "int32":
            return pa.int32()
        elif self.type_name == "int64":
            return pa.int64()
        elif self.type_name == "bool":
            return pa.bool_()
        elif self.type_name.startswith("image(") or self.type_name.startswith(
            "binary("
        ):
            # TODO we will need to think about how custom types work with tabular libraries
            return pa.binary()
        else:
            raise ValueError(f"Unsupported type conversion to pa: {self.type_name}")



import os

import pytest

from deltacat.experimental.storage.rivulet.dataset import Dataset
from deltacat.experimental.storage.rivulet.fs.file_store import FileStore
from deltacat.experimental.storage.rivulet.schema.datatype import Datatype
from deltacat.experimental.storage.rivulet.metastore.delta import DeltacatManifestIO
from deltacat.experimental.storage.rivulet import Schema, Field
import pyarrow as pa
import pyarrow.parquet


@pytest.fixture
def sample_schema():
    return Schema(
        fields=[
            Field("id", Datatype.int32(), is_merge_key=True),
            Field("name", Datatype.string()),
            Field("age", Datatype.int32()),
        ]
    )


@pytest.fixture
def sample_pydict():
    return {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"], "age": [25, 30, 35]}


@pytest.fixture
def path(tmp_path):
    return tmp_path


@pytest.fixture
def sample_parquet_data(path, sample_pydict):
    parquet_path = path / "test.parquet"
    table = pa.Table.from_pydict(sample_pydict)
    pyarrow.parquet.write_table(table, parquet_path)
    return parquet_path


def test_write_manifest_round_trip(sample_parquet_data, sample_schema):
    dataset = Dataset.from_parquet(
        file_uri=sample_parquet_data, name="dataset", merge_keys="id"
    )

    path, filesystem = FileStore.filesystem(dataset._metadata_path)
    file_store = FileStore(path, filesystem=filesystem)
    manifest_io = DeltacatManifestIO(path, dataset._locator)

    sst_files = ["sst1.sst", "sst2.sst"]
    schema = Schema(
        {("id", Datatype.int32()), ("name", Datatype.string())},
        "id",
    )
    level = 2

    uri = os.path.join(path, "manifest.json")

    file_store.create_output_file(uri)
    written = manifest_io.write(sst_files, schema, level)
    manifest = manifest_io.read(written)

    assert manifest.context.schema == schema
    assert manifest.context.level == level
    assert manifest.sst_files == sst_files
